@@ -5,14 +5,20 @@ import subprocess
 import logging
 import hashlib
 import re
+import sys
+import gzip
 import os
 import fnmatch
 import email.utils
 import email.policy
+import email.header
 import requests
 import urllib.parse
 import datetime
 import time
+import shutil
+import mailbox
+import pwd
 
 from pathlib import Path
 from tempfile import mkstemp
@@ -21,7 +27,7 @@ from email import charset
 charset.add_charset('utf-8', None)
 emlpolicy = email.policy.EmailPolicy(utf8=True, cte_type='8bit', max_line_length=None)
 
-__VERSION__ = '0.3.7'
+__VERSION__ = '0.4.0'
 ATTESTATION_FORMAT_VER = '0.1'
 
 logger = logging.getLogger('b4')
@@ -88,6 +94,12 @@ DEFAULT_CONFIG = {
     'attestation-checkmarks': 'fancy',
     # How long to keep things in cache before expiring (minutes)?
     'cache-expire': '10',
+    # Used when creating summaries for b4 ty
+    'thanks-commit-url-mask': None,
+    # See thanks-pr-template.example
+    'thanks-pr-template': None,
+    # See thanks-am-template.example
+    'thanks-am-template': None,
     # If this is not set, we'll use what we find in 
     # git-config for gpg.program, and if that's not set,
     # we'll use "gpg" and hope for the better
@@ -114,6 +126,7 @@ class LoreMailbox:
         self.msgid_map = dict()
         self.series = dict()
         self.covers = dict()
+        self.trailer_map = dict()
         self.followups = list()
         self.unknowns = list()
 
@@ -135,7 +148,72 @@ class LoreMailbox:
             return self.msgid_map[msgid]
         return None
 
-    def get_series(self, revision=None, sloppytrailers=False):
+    def backfill(self, revision):
+        if revision in self.covers and self.covers[revision] is not None:
+            patch = self.covers[revision]
+        else:
+            # Find first non-None member in patches
+            lser = self.series[revision]
+            patch = None
+            for patch in lser.patches:
+                if patch is not None:
+                    break
+        logger.info('---')
+        logger.info('Thread incomplete, attempting to backfill')
+        cachedir = get_cache_dir()
+        listmap = os.path.join(cachedir, 'lists.map.lookup')
+        if not os.path.exists(listmap):
+            # lists.map is a custom service running on lore.kernel.org, so it is
+            # meaningless to make this a configurable URL
+            session = get_requests_session()
+            resp = session.get('https://lore.kernel.org/lists.map')
+            if resp.status_code != 200:
+                logger.debug('Unable to retrieve lore.kernel.org/lists.map')
+                return
+            content = resp.content.decode('utf-8')
+            with open(listmap, 'w') as fh:
+                fh.write(content)
+        else:
+            with open(listmap, 'r') as fh:
+                content = fh.read()
+
+        projmap = dict()
+        for line in content.split('\n'):
+            if line.find(':') <= 0:
+                continue
+            chunks = line.split(':')
+            projmap[chunks[0]] = chunks[1].strip()
+
+        allto = email.utils.getaddresses([str(x) for x in patch.msg.get_all('to', [])])
+        allto += email.utils.getaddresses([str(x) for x in patch.msg.get_all('cc', [])])
+        listarc = patch.msg.get_all('list-archive', [])
+        for entry in allto:
+            if entry[1] in projmap:
+                projurl = 'https://lore.kernel.org/%s/' % projmap[entry[1]]
+                # Make sure we don't re-query the same project we just used
+                reused = False
+                for arcurl in listarc:
+                    if projurl in arcurl:
+                        reused = True
+                        break
+                if reused:
+                    continue
+                # Try to backfill from that project
+                tmp_mbox = mkstemp()[1]
+                get_pi_thread_by_msgid(patch.msgid, tmp_mbox, useproject=projmap[entry[1]])
+                mbx = mailbox.mbox(tmp_mbox)
+                was = len(self.msgid_map)
+                for msg in mbx:
+                    self.add_message(msg)
+                mbx.close()
+                os.unlink(tmp_mbox)
+                if len(self.msgid_map) > was:
+                    logger.info('Loaded %s messages from %s', len(self.msgid_map)-was, projurl)
+                if self.series[revision].complete:
+                    logger.info('Successfully backfilled missing patches')
+                    break
+
+    def get_series(self, revision=None, sloppytrailers=False, backfill=True):
         if revision is None:
             if not len(self.series):
                 return None
@@ -156,6 +234,9 @@ class LoreMailbox:
             logger.critical('All patches in series v%s are missing.', lser.revision)
             return None
 
+        if not lser.complete and backfill:
+            self.backfill(revision)
+
         # Grab our cover letter if we have one
         if revision in self.covers.keys():
             lser.add_patch(self.covers[revision])
@@ -179,13 +260,9 @@ class LoreMailbox:
             if not len(fmsg.trailers):
                 logger.debug('  no trailers found, skipping')
                 continue
-            # if it's for the wrong revision, ignore it
-            if not fmsg.revision_inferred and lser.revision != fmsg.revision:
-                logger.debug('  follow-up for the wrong revision, skipping')
-                continue
             # Go up through the follow-ups and tally up trailers until
             # we either run out of in-reply-tos, or we find a patch in
-            # our series
+            # one of our series
             if fmsg.in_reply_to is None:
                 # Check if there's something matching in References
                 refs = fmsg.msg.get('References', '')
@@ -210,27 +287,46 @@ class LoreMailbox:
                 logger.debug('%sTrailers:', ' ' * lvl)
                 for trailer in trailers:
                     logger.debug('%s%s: %s', ' ' * (lvl+1), trailer[0], trailer[1])
-                found = False
-                if lser.revision != pmsg.revision:
+                if pmsg.has_diff and not pmsg.reply:
+                    # We found the patch for these trailers
+                    if pmsg.revision != revision:
+                        # add this into our trailer map to carry over trailers from
+                        # previous revisions to current revision if patch/metadata did
+                        # not change
+                        pmsg.load_hashes()
+                        attid = pmsg.attestation.attid
+                        if attid not in self.trailer_map:
+                            self.trailer_map[attid] = set()
+                        self.trailer_map[attid].update(trailers)
+                    pmsg.followup_trailers.update(trailers)
                     break
-                for lmsg in lser.patches:
-                    if lmsg is not None and lmsg.msgid == pmsg.msgid:
-                        # Confirmed, this is our parent patch
-                        lmsg.followup_trailers.update(trailers)
-                        found = True
-                        break
-                if found:
+                if pmsg.has_diffstat and not pmsg.reply:
+                    # Could be a cover letter
+                    pmsg.followup_trailers.update(trailers)
                     break
-                elif pmsg.in_reply_to and pmsg.in_reply_to in self.msgid_map:
+                if pmsg.in_reply_to and pmsg.in_reply_to in self.msgid_map:
                     lvl += 1
                     trailers.update(pmsg.trailers)
                     pmsg = self.msgid_map[pmsg.in_reply_to]
-                else:
-                    break
+                    continue
+                break
+
+        # Carry over trailers from previous series if patch/metadata did not change
+        for lmsg in lser.patches:
+            if lmsg is None or lmsg.attestation is None:
+                continue
+            lmsg.load_hashes()
+            if lmsg.attestation.attid in self.trailer_map:
+                lmsg.followup_trailers.update(self.trailer_map[lmsg.attestation.attid])
 
         return lser
 
     def add_message(self, msg):
+        msgid = LoreMessage.get_clean_msgid(msg)
+        if msgid in self.msgid_map:
+            logger.debug('Already have a message with this msgid, skipping %s', msgid)
+            return
+
         lmsg = LoreMessage(msg)
         logger.debug('Looking at: %s', lmsg.full_subject)
         self.msgid_map[lmsg.msgid] = lmsg
@@ -273,7 +369,7 @@ class LoreMailbox:
             if lmsg.revision not in self.series:
                 self.series[lmsg.revision] = LoreSeries(lmsg.revision, lmsg.expected)
                 if len(self.series) > 1:
-                    logger.info('Found new series v%s', lmsg.revision)
+                    logger.debug('Found new series v%s', lmsg.revision)
 
             # Attempt to auto-number series from the same author who did not bother
             # to set v2, v3, etc in the patch revision
@@ -355,7 +451,7 @@ class LoreSeries:
             return 'undefined'
 
         prefix = lmsg.date.strftime('%Y%m%d')
-        authorline = email.utils.getaddresses(lmsg.msg.get_all('from', []))[0]
+        authorline = email.utils.getaddresses([str(x) for x in lmsg.msg.get_all('from', [])])[0]
         if extended:
             local = authorline[1].split('@')[0]
             unsafe = '%s_%s_%s' % (prefix, local, lmsg.subject)
@@ -437,7 +533,7 @@ class LoreSeries:
                             attpolicy = 'softfail'
                         # Make sure it's not too old compared to the message date
                         # Timezone doesn't matter as we calculate whole days
-                        tdelta = lmsg.date.replace(tzinfo=None) - attdoc.sigdate
+                        tdelta = lmsg.date.replace(tzinfo=None) - attdoc.lsig.sigdate
                         if tdelta.days > attstaled:
                             # Uh-oh, attestation is too old!
                             logger.info('  %s %s', attfail, lmsg.full_subject)
@@ -445,7 +541,7 @@ class LoreSeries:
                                                                                              attstaled, tdelta.days))
                         else:
                             logger.info('  %s %s', attpass, lmsg.full_subject)
-                            attdata[at-1] = attdoc.attestor.get_trailer(lmsg.fromemail)
+                            attdata[at-1] = attdoc.lsig.attestor.get_trailer(lmsg.fromemail)
                 else:
                     logger.info('  %s', lmsg.full_subject)
 
@@ -520,7 +616,16 @@ class LoreMessage:
         self.trailers = set()
         self.followup_trailers = set()
 
+        # These are populated by pr
+        self.pr_base_commit = None
+        self.pr_repo = None
+        self.pr_ref = None
+        self.pr_tip_commit = None
+        self.pr_remote_tip_commit = None
+
         self.attestation = None
+        # Patchwork hash
+        self.pwhash = None
 
         self.msgid = LoreMessage.get_clean_msgid(self.msg)
         self.lsubject = LoreSubject(msg['Subject'])
@@ -541,15 +646,17 @@ class LoreMessage:
         self.in_reply_to = LoreMessage.get_clean_msgid(self.msg, header='In-Reply-To')
 
         try:
-            fromdata = email.utils.getaddresses(self.msg.get_all('from', []))[0]
+            fromdata = email.utils.getaddresses([str(x) for x in self.msg.get_all('from', [])])[0]
             self.fromname = fromdata[0]
             self.fromemail = fromdata[1]
+            if not len(self.fromname.strip()):
+                self.fromname = self.fromemail
         except IndexError:
             pass
 
         self.date = email.utils.parsedate_to_datetime(str(self.msg['Date']))
 
-        diffre = re.compile(r'^(---.*\n\+\+\+|GIT binary patch)', re.M | re.I)
+        diffre = re.compile(r'^(---.*\n\+\+\+|GIT binary patch|diff --git \w/\S+ \w/\S+)', re.M | re.I)
         diffstatre = re.compile(r'^\s*\d+ file.*\d+ (insertion|deletion)', re.M | re.I)
 
         # walk until we find the first text/plain part
@@ -598,7 +705,7 @@ class LoreMessage:
             # Do we have something that looks like a person-trailer?
             matches = re.findall(r'^\s*([\w-]+):[ \t]+(.*<\S+>)\s*$', self.body, re.MULTILINE)
             # These are commonly part of patch/commit metadata
-            badtrailers = ('from', 'author', 'email')
+            badtrailers = ('from', 'author')
             if matches:
                 for tname, tvalue in matches:
                     if tname.lower() not in badtrailers:
@@ -611,7 +718,6 @@ class LoreMessage:
 
         trailers = set()
         for tname, tvalue in self.trailers:
-            # Don't do from-matching on certain trailers like Fixes:
             if tname.lower() in ('fixes',):
                 trailers.add((tname, tvalue))
                 continue
@@ -687,7 +793,21 @@ class LoreMessage:
 
     @staticmethod
     def clean_header(hdrval):
-        new_hdrval = re.sub(r'\n?\s+', ' ', str(hdrval))
+        if hdrval is None:
+            return ''
+
+        decoded = ''
+        for hstr, hcs in email.header.decode_header(hdrval):
+            if hcs is None:
+                hcs = 'utf-8'
+            try:
+                decoded += hstr.decode(hcs)
+            except LookupError:
+                # Try as utf-u
+                decoded += hstr.decode('utf-8', errors='replace')
+            except (UnicodeDecodeError, AttributeError):
+                decoded += hstr
+        new_hdrval = re.sub(r'\n?\s+', ' ', decoded)
         return new_hdrval.strip()
 
     @staticmethod
@@ -701,29 +821,99 @@ class LoreMessage:
         return msgid
 
     @staticmethod
-    def get_patch_hash(diff):
-        # The aim is to represent the patch as if you did the following:
-        # git diff HEAD~.. | dos2unix | sha256sum
-        #
-        # This subroutine removes anything at the beginning of diff data, like
-        # diffstat or any other auxiliary data, and anything trailing at the end
-        # XXX: This currently doesn't work for git binary patches
-        #
+    def get_patchwork_hash(diff):
+        # Make sure we just have the diff without any extraneous content.
+        diff = LoreMessage.get_clean_diff(diff)
+        """Generate a hash from a diff. Lifted verbatim from patchwork."""
+
+        prefixes = ['-', '+', ' ']
+        hashed = hashlib.sha1()
+
+        for line in diff.split('\n'):
+            if len(line) <= 0:
+                continue
+
+            hunk_match = HUNK_RE.match(line)
+            filename_match = FILENAME_RE.match(line)
+
+            if filename_match:
+                # normalise -p1 top-directories
+                if filename_match.group(1) == '---':
+                    filename = 'a/'
+                else:
+                    filename = 'b/'
+                filename += '/'.join(filename_match.group(2).split('/')[1:])
+
+                line = filename_match.group(1) + ' ' + filename
+            elif hunk_match:
+                # remove line numbers, but leave line counts
+                def fn(x):
+                    if not x:
+                        return 1
+                    return int(x)
+
+                line_nos = list(map(fn, hunk_match.groups()))
+                line = '@@ -%d +%d @@' % tuple(line_nos)
+            elif line[0] in prefixes:
+                # if we have a +, - or context line, leave as-is
+                pass
+            else:
+                # other lines are ignored
+                continue
+
+            hashed.update((line + '\n').encode('utf-8'))
+
+        return hashed.hexdigest()
+
+    @staticmethod
+    def get_clean_diff(diff):
         diff = diff.replace('\r', '')
 
         # For keeping a buffer of lines preceding @@ ... @@
         buflines = list()
-
-        phasher = hashlib.sha256()
+        difflines = ''
 
         # Used for counting where we are in the patch
-        pp = 0
+        pp = mm = 0
+        inside_binary_chunk = False
         for line in diff.split('\n'):
+            if not len(line):
+                if inside_binary_chunk:
+                    inside_binary_chunk = False
+                    # add all buflines to difflines
+                    difflines += '\n'.join(buflines) + '\n\n'
+                    buflines = list()
+                    continue
+                buflines.append(line)
+                continue
+            elif inside_binary_chunk:
+                buflines.append(line)
+                continue
+            # If line starts with 'index ' and previous line starts with 'deleted ', then
+            # it's a file delete and therefore doesn't have a regular hunk.
+            if line.find('index ') == 0 and len(buflines) > 1 and buflines[-1].find('deleted ') == 0:
+                # add this and 2 preceding lines to difflines and reset buflines
+                buflines.append(line)
+                difflines += '\n'.join(buflines[-3:]) + '\n'
+                buflines = list()
+                continue
+            if line.find('delta ') == 0 or line.find('literal ') == 0:
+                # we are inside a binary patch
+                inside_binary_chunk = True
+                buflines.append(line)
+                continue
             hunk_match = HUNK_RE.match(line)
             if hunk_match:
                 # logger.debug('Crunching %s', line)
                 mlines, plines = hunk_match.groups()
-                pp = int(plines)
+                try:
+                    pp = int(plines)
+                except TypeError:
+                    pp = 1
+                try:
+                    mm = int(mlines)
+                except TypeError:
+                    mm = 1
                 addlines = list()
                 for bline in reversed(buflines):
                     # Go backward and add lines until we get to the start
@@ -732,24 +922,39 @@ class LoreMessage:
                         break
                     addlines.append(bline)
                 if addlines:
-                    phasher.update(('\n'.join(reversed(addlines)) + '\n').encode('utf-8'))
+                    difflines += '\n'.join(reversed(addlines)) + '\n'
                 buflines = list()
                 # Feed this line to the hasher
-                phasher.update((line + '\n').encode('utf-8'))
+                difflines += line + '\n'
                 continue
-            if pp > 0:
+            if pp > 0 or mm > 0:
                 # Inside the patch
-                phasher.update((line + '\n').encode('utf-8'))
-                if len(line) and line[0] == '-':
-                    continue
-                pp -= 1
+                difflines += line + '\n'
+                if line[0] in (' ', '-'):
+                    mm -= 1
+                if line[0] in (' ', '+'):
+                    pp -= 1
                 continue
             # Not anything we recognize, so stick into buflines
             buflines.append(line)
+        return difflines
 
+    @staticmethod
+    def get_patch_hash(diff):
+        # The aim is to represent the patch as if you did the following:
+        # git diff HEAD~.. | dos2unix | sha256sum
+        #
+        # This subroutine removes anything at the beginning of diff data, like
+        # diffstat or any other auxiliary data, and anything trailing at the end
+        #
+        diff = LoreMessage.get_clean_diff(diff)
+        phasher = hashlib.sha256()
+        phasher.update(diff.encode('utf-8'))
         return phasher.hexdigest()
 
     def load_hashes(self):
+        if self.attestation is not None:
+            return
         msg_out = mkstemp()
         patch_out = mkstemp()
         cmdargs = ['mailinfo', '--encoding=UTF-8', msg_out[1], patch_out[1]]
@@ -780,6 +985,7 @@ class LoreMessage:
             patch = pfh.read()
             if len(patch.strip()):
                 p = LoreMessage.get_patch_hash(patch)
+                self.pwhash = LoreMessage.get_patchwork_hash(patch)
         os.unlink(patch_out[1])
 
         if i and m and p:
@@ -870,7 +1076,7 @@ class LoreMessage:
                 # We return the first hit
                 return attdoc
             # Does this doc have an exact match?
-            uid = attdoc.attestor.get_matching_uid(self.fromemail)
+            uid = attdoc.lsig.attestor.get_matching_uid(self.fromemail)
             if uid[1] == self.fromemail:
                 return attdoc
             # stick an error in the first available attdoc saying
@@ -1016,15 +1222,63 @@ class LoreAttestor:
         return '\n'.join(out)
 
 
-class LoreAttestationDocument:
-    def __init__(self, source, sigdata):
-        self.source = source
+class LoreAttestationSignature:
+    def __init__(self, output, trustmodel):
         self.good = False
         self.valid = False
         self.trusted = False
         self.sigdate = None
-        self.passing = False
         self.attestor = None
+        self.errors = set()
+
+        gs_matches = re.search(r'^\[GNUPG:\] GOODSIG ([0-9A-F]+)\s+.*$', output, re.M)
+        if gs_matches:
+            logger.debug('  GOODSIG')
+            self.good = True
+            keyid = gs_matches.groups()[0]
+            self.attestor = LoreAttestor(keyid)
+            puid = '%s <%s>' % self.attestor.get_primary_uid()
+            vs_matches = re.search(r'^\[GNUPG:\] VALIDSIG ([0-9A-F]+) (\d{4}-\d{2}-\d{2}) (\d+)', output, re.M)
+            if vs_matches:
+                logger.debug('  VALIDSIG')
+                self.valid = True
+                ymd = vs_matches.groups()[1]
+                self.sigdate = datetime.datetime.strptime(ymd, '%Y-%m-%d')
+                # Do we have a TRUST_(FULLY|ULTIMATE)?
+                ts_matches = re.search(r'^\[GNUPG:\] TRUST_(FULLY|ULTIMATE)', output, re.M)
+                if ts_matches:
+                    logger.debug('  TRUST_%s', ts_matches.groups()[0])
+                    self.trusted = True
+                else:
+                    self.errors.add('Insufficient trust (model=%s): %s (%s)'
+                                    % (trustmodel, keyid, puid))
+            else:
+                self.errors.add('Signature not valid from key: %s (%s)' % (keyid, puid))
+        else:
+            # Are we missing a key?
+            matches = re.search(r'^\[GNUPG:\] NO_PUBKEY ([0-9A-F]+)$', output, re.M)
+            if matches:
+                self.errors.add('Missing public key: %s' % matches.groups()[0])
+
+    def __repr__(self):
+        out = list()
+        out.append('  good: %s' % self.good)
+        out.append('  valid: %s' % self.valid)
+        out.append('  trusted: %s' % self.trusted)
+        if self.attestor is not None:
+            out.append('  attestor: %s' % self.attestor.keyid)
+
+        out.append('  --- validation errors ---')
+        for error in self.errors:
+            out.append('  | %s' % error)
+        return '\n'.join(out)
+
+
+class LoreAttestationDocument:
+    def __init__(self, source, sigdata):
+        self.source = source
+        self.lsig = None
+        self.passing = False
         self.hashes = set()
         self.errors = set()
 
@@ -1035,42 +1289,14 @@ class LoreAttestationDocument:
 
         logger.debug('Validating document obtained from %s', self.source)
         ecode, output = gpg_run_command(gpgargs, stdin=sigdata.encode('utf-8'))
-        if ecode == 0:
-            # We're looking for both GOODSIG and VALIDSIG
-            gs_matches = re.search(r'^\[GNUPG:\] GOODSIG ([0-9A-F]+)\s+.*$', output, re.M)
-            if gs_matches:
-                logger.debug('  GOODSIG')
-                self.good = True
-                keyid = gs_matches.groups()[0]
-                self.attestor = LoreAttestor(keyid)
-                puid = '%s <%s>' % self.attestor.get_primary_uid()
-                vs_matches = re.search(r'^\[GNUPG:\] VALIDSIG ([0-9A-F]+) (\d{4}-\d{2}-\d{2}) (\d+)', output, re.M)
-                if vs_matches:
-                    logger.debug('  VALIDSIG')
-                    self.valid = True
-                    ymd = vs_matches.groups()[1]
-                    self.sigdate = datetime.datetime.strptime(ymd, '%Y-%m-%d')
-                    # Do we have a TRUST_(FULLY|ULTIMATE)?
-                    ts_matches = re.search(r'^\[GNUPG:\] TRUST_(FULLY|ULTIMATE)', output, re.M)
-                    if ts_matches:
-                        logger.debug('  TRUST_%s', ts_matches.groups()[0])
-                        self.trusted = True
-                    else:
-                        self.errors.add('Insufficient trust (model=%s): %s (%s)'
-                                        % (config['attestation-trust-model'], keyid, puid))
-                else:
-                    self.errors.add('Signature not valid from key: %s (%s)' % (keyid, puid))
-        else:
-            # Are we missing a key?
-            matches = re.search(r'^\[GNUPG:\] NO_PUBKEY ([0-9A-F]+)$', output, re.M)
-            if matches:
-                self.errors.add('Missing public key: %s' % matches.groups()[0])
-            else:
-                logger.debug('NOGOOD: Signature on %s failed to verify', self.source)
-            return
+        self.lsig = LoreAttestationSignature(output, config['attestation-trust-model'])
+        self.errors.update(self.lsig.errors)
 
-        if self.good and self.valid and self.trusted:
+        if self.lsig.good and self.lsig.valid and self.lsig.trusted:
             self.passing = True
+        else:
+            # Not going any further
+            return
 
         if source.find('http') == 0:
             # We only cache known-good attestations obtained from remote
@@ -1103,19 +1329,14 @@ class LoreAttestationDocument:
     def __repr__(self):
         out = list()
         out.append('  source: %s' % self.source)
-        out.append('  good: %s' % self.good)
-        out.append('  valid: %s' % self.valid)
-        out.append('  trusted: %s' % self.trusted)
-        if self.attestor is not None:
-            out.append('  attestor: %s' % self.attestor.keyid)
-
         out.append('  --- validation errors ---')
         for error in self.errors:
             out.append('  | %s' % error)
         out.append('  --- hashes ---')
         for hg in self.hashes:
             out.append('  | %s-%s-%s' % (hg[0][:8], hg[1][:8], hg[2][:8]))
-        return '\n'.join(out)
+        ret = '\n'.join(out) + '\n' + str(self.lsig)
+        return ret
 
     @staticmethod
     def get_from_cache(attid):
@@ -1243,7 +1464,9 @@ def _run_command(cmdargs, stdin=None, logstderr=False):
     output = output.decode('utf-8', errors='replace')
 
     if logstderr and len(error.strip()):
-        logger.debug('Stderr: %s', error.decode('utf-8', errors='replace'))
+        errout = error.decode('utf-8', errors='replace')
+        logger.debug('Stderr: %s', errout)
+        output += errout
 
     return sp.returncode, output
 
@@ -1261,6 +1484,8 @@ def gpg_run_command(args, stdin=None, logstderr=False):
 def git_run_command(gitdir, args, stdin=None, logstderr=False):
     cmdargs = ['git', '--no-pager']
     if gitdir:
+        if os.path.isdir(os.path.join(gitdir, '.git')):
+            gitdir = os.path.join(gitdir, '.git')
         cmdargs += ['--git-dir', gitdir]
     cmdargs += args
 
@@ -1316,6 +1541,16 @@ def get_main_config():
     return MAIN_CONFIG
 
 
+def get_data_dir():
+    if 'XDG_DATA_HOME' in os.environ:
+        datahome = os.environ['XDG_DATA_HOME']
+    else:
+        datahome = os.path.join(str(Path.home()), '.local', 'share')
+    datadir = os.path.join(datahome, 'b4')
+    Path(datadir).mkdir(parents=True, exist_ok=True)
+    return datadir
+
+
 def get_cache_dir():
     global _CACHE_CLEANED
     if 'XDG_CACHE_HOME' in os.environ:
@@ -1350,6 +1585,9 @@ def get_user_config():
     global USER_CONFIG
     if USER_CONFIG is None:
         USER_CONFIG = get_config_from_git(r'user\..*')
+        if 'name' not in USER_CONFIG:
+            udata = pwd.getpwuid(os.getuid())
+            USER_CONFIG['name'] = udata.pw_gecos
     return USER_CONFIG
 
 
@@ -1361,6 +1599,40 @@ def get_requests_session():
     return REQSESSION
 
 
+def get_msgid_from_stdin():
+    if not sys.stdin.isatty():
+        message = email.message_from_string(sys.stdin.read())
+        return message.get('Message-ID', None)
+    logger.error('Error: pipe a message or pass msgid as parameter')
+    sys.exit(1)
+
+
+def get_msgid(cmdargs):
+    if not cmdargs.msgid:
+        logger.debug('Getting Message-ID from stdin')
+        msgid = get_msgid_from_stdin()
+        if msgid is None:
+            logger.error('Unable to find a valid message-id in stdin.')
+            sys.exit(1)
+    else:
+        msgid = cmdargs.msgid
+
+    msgid = msgid.strip('<>')
+    # Handle the case when someone pastes a full URL to the message
+    matches = re.search(r'^https?://[^/]+/([^/]+)/([^/]+@[^/]+)', msgid, re.IGNORECASE)
+    if matches:
+        chunks = matches.groups()
+        msgid = chunks[1]
+        # Infer the project name from the URL, if possible
+        if chunks[0] != 'r':
+            cmdargs.useproject = chunks[0]
+    # Handle special case when msgid is prepended by id: or rfc822msgid:
+    if msgid.find('id:') >= 0:
+        msgid = re.sub(r'^\w*id:', '', msgid)
+
+    return msgid
+
+
 def save_strict_thread(in_mbx, out_mbx, msgid):
     want = {msgid}
     got = set()
@@ -1410,3 +1682,109 @@ def save_strict_thread(in_mbx, out_mbx, msgid):
 
     if len(in_mbx) > len(out_mbx):
         logger.info('Reduced thread to strict matches only (%s->%s)', len(in_mbx), len(out_mbx))
+
+
+def get_pi_thread_by_url(t_mbx_url, savefile):
+    session = get_requests_session()
+    resp = session.get(t_mbx_url)
+    if resp.status_code != 200:
+        logger.critical('Server returned an error: %s', resp.status_code)
+        return None
+    t_mbox = gzip.decompress(resp.content)
+    resp.close()
+    if not len(t_mbox):
+        logger.critical('No messages found for that query')
+        return None
+    with open(savefile, 'wb') as fh:
+        logger.debug('Saving %s', savefile)
+        fh.write(t_mbox)
+    return savefile
+
+
+def get_pi_thread_by_msgid(msgid, savefile, useproject=None, nocache=False):
+    qmsgid = urllib.parse.quote_plus(msgid)
+    config = get_main_config()
+    cachedir = get_cache_dir()
+    base = msgid
+    if useproject:
+        base = '%s-%s' % (useproject, msgid)
+    cachefile = os.path.join(cachedir, '%s.pi.mbx' % urllib.parse.quote_plus(base))
+    if os.path.exists(cachefile) and not nocache:
+        logger.debug('Using cached copy: %s', cachefile)
+        shutil.copyfile(cachefile, savefile)
+        return savefile
+
+    # Grab the head from lore, to see where we are redirected
+    midmask = config['midmask'] % qmsgid
+    loc = urllib.parse.urlparse(midmask)
+    if useproject:
+        projurl = '%s://%s/%s' % (loc.scheme, loc.netloc, useproject)
+    else:
+        logger.info('Looking up %s', midmask)
+        session = get_requests_session()
+        resp = session.head(midmask)
+        if resp.status_code < 300 or resp.status_code > 400:
+            logger.critical('That message-id is not known.')
+            return None
+        # Pop msgid from the end of the redirect
+        chunks = resp.headers['Location'].rstrip('/').split('/')
+        projurl = '/'.join(chunks[:-1])
+        resp.close()
+    t_mbx_url = '%s/%s/t.mbox.gz' % (projurl, qmsgid)
+    logger.debug('t_mbx_url=%s', t_mbx_url)
+
+    logger.critical('Grabbing thread from %s', projurl.split('://')[1])
+    in_mbxf = get_pi_thread_by_url(t_mbx_url, '%s-loose' % savefile)
+    if not in_mbxf:
+        return None
+    in_mbx = mailbox.mbox(in_mbxf)
+    out_mbx = mailbox.mbox(savefile)
+    save_strict_thread(in_mbx, out_mbx, msgid)
+    in_mbx.close()
+    out_mbx.close()
+    os.unlink(in_mbxf)
+    shutil.copyfile(savefile, cachefile)
+    return savefile
+
+
+def git_format_patches(gitdir, start, end, reroll=None):
+    gitargs = ['format-patch', '--stdout']
+    if reroll is not None:
+        gitargs += ['-v', str(reroll)]
+    gitargs += ['%s..%s' % (start, end)]
+    ecode, out = git_run_command(gitdir, gitargs)
+    return ecode, out
+
+
+def git_commit_exists(gitdir, commit_id):
+    gitargs = ['cat-file', '-e', commit_id]
+    ecode, out = git_run_command(gitdir, gitargs)
+    return ecode == 0
+
+
+def git_branch_contains(gitdir, commit_id):
+    gitargs = ['branch', '--format=%(refname:short)', '--contains', commit_id]
+    lines = git_get_command_lines(gitdir, gitargs)
+    return lines
+
+
+def format_addrs(pairs):
+    addrs = set()
+    for pair in pairs:
+        # Remove any quoted-printable header junk from the name
+        addrs.add(email.utils.formataddr((LoreMessage.clean_header(pair[0]), LoreMessage.clean_header(pair[1]))))
+    return ', '.join(addrs)
+
+
+def make_quote(body, maxlines=5):
+    quotelines = list()
+    qcount = 0
+    for line in body.split('\n'):
+        # Quote the first paragraph only and then [snip] if we quoted more than maxlines
+        if qcount > maxlines and (not len(line.strip()) or line.strip().find('---') == 0):
+            quotelines.append('> ')
+            quotelines.append('> [...]')
+            break
+        quotelines.append('> %s' % line.strip('\r\n'))
+        qcount += 1
+    return '\n'.join(quotelines)
