@@ -13,11 +13,10 @@ import email.message
 import email.utils
 import re
 import time
-import shutil
+import json
 
 import urllib.parse
 import xml.etree.ElementTree
-import gzip
 
 import b4
 
@@ -26,81 +25,8 @@ from tempfile import mkstemp
 logger = b4.logger
 
 
-def get_msgid_from_stdin():
-    if not sys.stdin.isatty():
-        message = email.message_from_string(sys.stdin.read())
-        return message.get('Message-ID', None)
-    logger.error('Error: pipe a message or pass msgid as parameter')
-    sys.exit(1)
-
-
-def get_pi_thread_by_url(t_mbx_url, savefile):
-    session = b4.get_requests_session()
-    resp = session.get(t_mbx_url)
-    if resp.status_code != 200:
-        logger.critical('Server returned an error: %s', resp.status_code)
-        return None
-    t_mbox = gzip.decompress(resp.content)
-    resp.close()
-    if not len(t_mbox):
-        logger.critical('No messages found for that query')
-        return None
-    with open(savefile, 'wb') as fh:
-        logger.debug('Saving %s', savefile)
-        fh.write(t_mbox)
-    return savefile
-
-
-def get_pi_thread_by_msgid(msgid, config, cmdargs):
-    wantname = cmdargs.wantname
-    outdir = cmdargs.outdir
-    if wantname:
-        savefile = os.path.join(outdir, wantname)
-    else:
-        # Save it into msgid.mbox
-        savefile = '%s.t.mbx' % msgid
-        savefile = os.path.join(outdir, savefile)
-
-    cachedir = b4.get_cache_dir()
-    cachefile = os.path.join(cachedir, '%s.pi.mbx' % urllib.parse.quote_plus(msgid))
-    if os.path.exists(cachefile) and not cmdargs.nocache:
-        logger.debug('Using cached copy: %s', cachefile)
-        shutil.copyfile(cachefile, savefile)
-        return savefile
-
-    # Grab the head from lore, to see where we are redirected
-    midmask = config['midmask'] % msgid
-    logger.info('Looking up %s', midmask)
-    session = b4.get_requests_session()
-    resp = session.head(midmask)
-    if resp.status_code < 300 or resp.status_code > 400:
-        logger.critical('That message-id is not known.')
-        return None
-    canonical = resp.headers['Location'].rstrip('/')
-    resp.close()
-    t_mbx_url = '%s/t.mbox.gz' % canonical
-
-    loc = urllib.parse.urlparse(t_mbx_url)
-    if cmdargs.useproject:
-        logger.debug('Modifying query to use %s', cmdargs.useproject)
-        t_mbx_url = '%s://%s/%s/%s/t.mbox.gz' % (
-            loc.scheme, loc.netloc, cmdargs.useproject, msgid)
-        logger.debug('Will query: %s', t_mbx_url)
-    logger.critical('Grabbing thread from %s', loc.netloc)
-    in_mbxf = get_pi_thread_by_url(t_mbx_url, '%s-loose' % savefile)
-    if not in_mbxf:
-        return None
-    in_mbx = mailbox.mbox(in_mbxf)
-    out_mbx = mailbox.mbox(savefile)
-    b4.save_strict_thread(in_mbx, out_mbx, msgid)
-    in_mbx.close()
-    out_mbx.close()
-    os.unlink(in_mbxf)
-    shutil.copyfile(savefile, cachefile)
-    return savefile
-
-
-def mbox_to_am(mboxfile, config, cmdargs):
+def mbox_to_am(mboxfile, cmdargs):
+    config = b4.get_main_config()
     outdir = cmdargs.outdir
     wantver = cmdargs.wantver
     wantname = cmdargs.wantname
@@ -113,7 +39,7 @@ def mbox_to_am(mboxfile, config, cmdargs):
     for key, msg in mbx.items():
         lmbx.add_message(msg)
 
-    lser = lmbx.get_series(revision=wantver)
+    lser = lmbx.get_series(revision=wantver, sloppytrailers=cmdargs.sloppytrailers)
     if lser is None and wantver is None:
         logger.critical('No patches found.')
         return
@@ -156,6 +82,13 @@ def mbox_to_am(mboxfile, config, cmdargs):
         for trailer in lser.patches[0].followup_trailers:
             logger.critical('      %s: %s', trailer[0], trailer[1])
         logger.critical('NOTE: Rerun with -t to apply them to all patches')
+    if len(lser.trailer_mismatches):
+        logger.critical('---')
+        logger.critical('NOTE: some trailers ignored due to from/email mismatches:')
+        for tvalue, fname, femail in lser.trailer_mismatches:
+            logger.critical('    ! Trailer: %s', tvalue)
+            logger.critical('         From: %s <%s>', fname, femail)
+        logger.critical('NOTE: Rerun with -S to apply them anyway')
 
     logger.critical('---')
     if not lser.complete:
@@ -203,10 +136,58 @@ def mbox_to_am(mboxfile, config, cmdargs):
         logger.critical('       git am %s', am_filename)
 
     am_mbx.close()
+    thanks_record_am(lser)
 
     return am_filename
 
 
+def thanks_record_am(lser):
+    if not lser.complete:
+        logger.debug('Incomplete series, not tracking for thanks')
+        return
+
+    # Are we tracking this already?
+    datadir = b4.get_data_dir()
+    slug = lser.get_slug(extended=True)
+    filename = '%s.am' % slug
+    # Check if we're tracking it already
+    for entry in os.listdir(datadir):
+        if entry == filename:
+            return
+
+    patches = list()
+    for pmsg in lser.patches[1:]:
+        pmsg.load_hashes()
+        if pmsg.attestation is None:
+            logger.debug('Unable to get hashes for all patches, not tracking for thanks')
+            return
+        patches.append((pmsg.subject, pmsg.pwhash))
+
+    lmsg = lser.patches[0]
+    if lmsg is None:
+        lmsg = lser.patches[1]
+
+    allto = email.utils.getaddresses([str(x) for x in lmsg.msg.get_all('to', [])])
+    allcc = email.utils.getaddresses([str(x) for x in lmsg.msg.get_all('cc', [])])
+
+    out = {
+        'msgid': lmsg.msgid,
+        'subject': lmsg.full_subject,
+        'fromname': lmsg.fromname,
+        'fromemail': lmsg.fromemail,
+        'to': b4.format_addrs(allto),
+        'cc': b4.format_addrs(allcc),
+        'references': b4.LoreMessage.clean_header(lmsg.msg['References']),
+        'sentdate': b4.LoreMessage.clean_header(lmsg.msg['Date']),
+        'quote': b4.make_quote(lmsg.body, maxlines=5),
+        'patches': patches,
+    }
+    fullpath = os.path.join(datadir, filename)
+    with open(fullpath, 'w', encoding='utf-8') as fh:
+        json.dump(out, fh, ensure_ascii=False, indent=4)
+        logger.debug('Wrote %s for thanks tracking', filename)
+
+
 def am_mbox_to_quilt(am_mbx, q_dirname):
     if os.path.exists(q_dirname):
         logger.critical('ERROR: Directory %s exists, not saving quilt patches', q_dirname)
@@ -335,7 +316,7 @@ def get_newest_series(mboxfile):
             continue
         t_mbx_url = '%st.mbox.gz' % link
         savefile = mkstemp('b4-get')[1]
-        nt_mboxfile = get_pi_thread_by_url(t_mbx_url, savefile)
+        nt_mboxfile = b4.get_pi_thread_by_url(t_mbx_url, savefile)
         nt_mbx = mailbox.mbox(nt_mboxfile)
         # Append all of these to the existing mailbox
         new_adds = 0
@@ -364,29 +345,18 @@ def main(cmdargs):
         # Force nocache mode
         cmdargs.nocache = True
 
-    config = b4.get_main_config()
-
     if not cmdargs.localmbox:
-        if not cmdargs.msgid:
-            logger.debug('Getting Message-ID from stdin')
-            msgid = get_msgid_from_stdin()
-            if msgid is None:
-                logger.error('Unable to find a valid message-id in stdin.')
-                sys.exit(1)
+        msgid = b4.get_msgid(cmdargs)
+        wantname = cmdargs.wantname
+        outdir = cmdargs.outdir
+        if wantname:
+            savefile = os.path.join(outdir, wantname)
         else:
-            msgid = cmdargs.msgid
+            # Save it into msgid.mbox
+            savefile = '%s.t.mbx' % msgid
+            savefile = os.path.join(outdir, savefile)
 
-        msgid = msgid.strip('<>')
-        # Handle the case when someone pastes a full URL to the message
-        matches = re.search(r'^https?://[^/]+/([^/]+)/([^/]+@[^/]+)', msgid, re.IGNORECASE)
-        if matches:
-            chunks = matches.groups()
-            msgid = chunks[1]
-            # Infer the project name from the URL, if possible
-            if chunks[0] != 'r':
-                cmdargs.useproject = chunks[0]
-
-        mboxfile = get_pi_thread_by_msgid(msgid, config, cmdargs)
+        mboxfile = b4.get_pi_thread_by_msgid(msgid, savefile, useproject=cmdargs.useproject, nocache=cmdargs.nocache)
         if mboxfile is None:
             return
 
@@ -404,7 +374,7 @@ def main(cmdargs):
         get_newest_series(threadmbox)
 
     if cmdargs.subcmd == 'am':
-        mbox_to_am(threadmbox, config, cmdargs)
+        mbox_to_am(threadmbox, cmdargs)
         if not cmdargs.localmbox:
             os.unlink(threadmbox)
     else:
